/*
 * Constructs dynamic checks for things like runtime stack checks
 */

package rewrite.pointcut;

import rewrite.pointcut.PointcutData.*;
import rewrite.util.*;

import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.Repository;
import org.apache.bcel.util.*;
import org.apache.bcel.Constants;

import spox.policy.pointcut.PointcutComponent.*;

/**
 * Written by Micah Jones
 */
public class DynamicCodeBuilder
{
    
    // Methods
    public static InstructionBlock mergeInstructionBlockTree(
            InstructionBlock iBlock)
    {
        // Merges a tree of instruction blocks into one instruction list
        // and assigns that list to pData.returnIlist
        // Note that this function is recursive
        
        // If the bopContainer field is null, we're at a base case
        if (iBlock.bopContainer == null)
            return iBlock;
        
        InstructionList ilist = iBlock.ilist;
        
        // Which kind of boolean op are we contained within?
        switch (iBlock.bopContainer)
        {
            case and:
            {
                for (int i = 0; i < iBlock.iBlockArray.size(); i++)
                {
                    // First make sure this iBlock is merged with its children
                    iBlock.iBlockArray.set(i,
                            mergeInstructionBlockTree(
                                iBlock.iBlockArray.get(i)));
                    // Since the last instruction in a list corresponds
                    // to a false evaluation, we insert gotos around that
                    // Before, a jump to our iBlock's first instruction
                    // After, a jump to our iBlock's last one (false branch)
                    // Note also that we're inserting the underlying instruction
                    // list ABOVE our current instruction set
                    if (iBlock.iBlockArray.get(i).ilist.size() > 2)
                    {
                        InstructionHandle falseIh = ilist.getEnd();
                        InstructionHandle trueIh = ilist.getStart();
                        ilist.insert(iBlock.iBlockArray.get(i).ilist);
                        InstructionHandle falseIh2 = trueIh.getPrev();
                        InstructionHandle trueIh2 = falseIh2.getPrev();
                        ilist.insert(falseIh2,
                                InstructionFactory.createBranchInstruction(
                                    Constants.GOTO, trueIh));
                        ilist.append(falseIh2,
                                InstructionFactory.createBranchInstruction(
                                    Constants.GOTO, falseIh));
                        try
                        {
//                        ilist.delete(trueIh2);
//                        ilist.delete(falseIh2);
                        } catch (Exception ex) {}
                    }
                }
                break;
            }
            case or:
            {
                for (int i = 0; i < iBlock.iBlockArray.size(); i++)
                {
                    // First make sure this iBlock is merged with its children
                    iBlock.iBlockArray.set(i,
                            mergeInstructionBlockTree(
                                iBlock.iBlockArray.get(i)));
                    // The last instruction corresponds to false, the one
                    // before it corresponds to true
                    // False means we let the instructions fall through to
                    // ours (unless we're actually en empty 2-NOP list)
                    // True means we create a jump to OUR true branch
                    if (iBlock.iBlockArray.get(i).ilist.size() > 2)
                    {
                        InstructionHandle oldStartIh = ilist.getStart();
                        InstructionHandle falseIh;
                        if (iBlock.ilist.size() > 2)
                            falseIh = ilist.getStart();
                        else
                            falseIh = ilist.getEnd();
                        InstructionHandle trueIh = ilist.getEnd().getPrev();
                        ilist.insert(iBlock.iBlockArray.get(i).ilist);
                        InstructionHandle falseIh2 = oldStartIh.getPrev();
                        InstructionHandle trueIh2 = falseIh2.getPrev();
                        ilist.insert(falseIh2,
                                InstructionFactory.createBranchInstruction(
                                    Constants.GOTO, trueIh));
                        if (falseIh == ilist.getEnd())
                        {
                            // Only make a jump if there's no comparisons left
                            ilist.append(falseIh2,
                                InstructionFactory.createBranchInstruction(
                                    Constants.GOTO, falseIh));
                        }
                        try
                        {
//                        ilist.delete(trueIh2);
//                        ilist.delete(falseIh2);
                        } catch (Exception ex) {}
                    }
                }
                break;
            }
            case not:
            {
                // A not really should only contain exactly one iBlock, but
                // we use a for loop here to be safe
                for (int i = 0; i < iBlock.iBlockArray.size(); i++)
                {
                    // First make sure this iBlock is merged with its children
                    iBlock.iBlockArray.set(i,
                            mergeInstructionBlockTree(
                                iBlock.iBlockArray.get(i)));
                    // The last instruction corresponds to false, the one
                    // before it corresponds to true
                    // Since this is merely a not, we invert the destinations
                    // when assigning jumps to our own branches
                    if (iBlock.iBlockArray.get(i).ilist.size() > 2)
                    {
                        InstructionHandle falseIh = ilist.getEnd();
                        InstructionHandle trueIh = ilist.getStart();
                        ilist.insert(iBlock.iBlockArray.get(i).ilist);
                        InstructionHandle falseIh2 = trueIh.getPrev();
                        InstructionHandle trueIh2 = falseIh2.getPrev();
                        ilist.insert(falseIh2,
                                InstructionFactory.createBranchInstruction(
                                    Constants.GOTO, falseIh));
                        ilist.append(falseIh2,
                                InstructionFactory.createBranchInstruction(
                                    Constants.GOTO, trueIh));
                        try
                        {
//                        ilist.delete(trueIh2);
//                        ilist.delete(falseIh2);
                        } catch (Exception ex) {}
                    }
                }
                break;
            }
        }
        
        // Get the updated ilist and return the iBlock
        iBlock.ilist = ilist;
        return iBlock;
        
    }

    public static void buildArgValCheck(
            ArgVal argVal,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] type)
    {
        
        // Handle differently depending on what sort of instruction
        // we're working with
        if (pData.instruction == null)
            buildArgValCheckExecution(argVal, pData, iBlock, type);
        else if (pData.instruction instanceof InvokeInstruction ||
                 pData.instruction instanceof ReturnInstruction)
            buildArgValCheckInvoke(argVal, pData, iBlock, type);
        else if (pData.instruction instanceof FieldInstruction &&
                 pData.instruction.getName().startsWith("get"))
            buildArgValCheckGetField(argVal, pData, iBlock, type);
        else if (pData.instruction instanceof FieldInstruction)
            buildArgValCheckPutField(argVal, pData, iBlock, type);

    }

    private static void buildArgValCheckExecution(
            ArgVal argVal,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] type)
    {
        
        // Locals
        LocalVariableGen[] lvgen;
        InstructionFactory ifact = new InstructionFactory(pData.cgen);
        
        // Initialize the new InstructionList
        InstructionList ilist = iBlock.ilist;
        
        // Obtain local variables (the parameters are in the first n or
        // n+1 variables, depending on whether or not this method is static)
        lvgen = pData.mgen.getLocalVariables();
        int offset = 1;
        if ((pData.cgen.getModifiers() & Constants.ACC_STATIC) != 0)
            offset = 0;
        
        // False branches eventually jump to the nop at the end of the list
        InstructionHandle falseIh = ilist.getEnd();
        InstructionHandle ih = falseIh;
        
        // If we're within an OR, we need to insert a new nop BEFORE the
        // false position to act as the truth branch (if there isn't one
        // there already)
        InstructionHandle trueIh = ih;
        if (iBlock.bopContainer == BopContainer.or)
        {
            if (ilist.getLength() >= 2)
            {
                if (!(falseIh.getPrev().getInstruction() instanceof NOP))
                    trueIh = ilist.insert(falseIh, ifact.NOP);
                else
                    trueIh = falseIh.getPrev();
            }
            else
            {
                trueIh = ilist.insert(falseIh, ifact.NOP);
            }
            // Realign our insertion point to precede the true branch
            ih = trueIh;
            // Also, if the last non-nop instruction was a branch pointing to
            // the false branch, we need to flip it and redirect the jump to the
            // truth branch -- thus a false implies that we fall through to the
            // new check that we are currently inserting
            if (ilist.getLength() >= 3)
            {
                InstructionHandle tempIh = trueIh.getPrev();
                if (tempIh.getInstruction() instanceof IfInstruction)
                {
                    IfInstruction falseBranch =
                            (IfInstruction)tempIh.getInstruction();
                    if (falseBranch.containsTarget(falseIh))
                    {
                        falseBranch = falseBranch.negate();
                        falseBranch.setTarget(trueIh);
                        tempIh.setInstruction(falseBranch);
                    }
                }
            }
        }
        
        if (argVal.valExpr instanceof IntCmp)
        {
            // Now push the value for comparison
            ilist.insert(ih, ifact.createLoad(type[argVal.num-1],
                    lvgen[argVal.num-1+offset].getIndex()));
            ilist.insert(ih, new PUSH(pData.cpgen, ((IntCmp)argVal.valExpr).n));
            // The if is dynamic to the type of the expression in the policy
            if (argVal.valExpr instanceof IntNE)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPEQ, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPNE, falseIh));
                }
            }
            else if (argVal.valExpr instanceof IntLT)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPGE, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPLT, falseIh));
                }
            }
            else if (argVal.valExpr instanceof IntGT)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPLE, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPGT, falseIh));
                }
            }
            else if (argVal.valExpr instanceof IntLE)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPGT, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPLE, falseIh));
                }
            }
            else if (argVal.valExpr instanceof IntGE)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPLT, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPGE, falseIh));
                }
            }
            else // IntEQ
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPNE, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPEQ, falseIh));
                }
            }
        }
        else if (argVal.valExpr instanceof StrEq)
        {
            // Create a regex comparison operation (Load the regular expression
            // first, then the input string, since the method takes them
            // in that order)
            ilist.insert(ih, new PUSH(pData.cpgen,
                    ((StrEq)argVal.valExpr).regex));
            // Obtain the result of toString()
            short kind = Constants.INVOKEVIRTUAL;
            try {
                JavaClass jclass = Repository.lookupClass(type[argVal.num-1].toString());
                if (jclass.isInterface())
                    kind = Constants.INVOKEINTERFACE;
            } catch (ClassNotFoundException ex) {/*Ignore for now*/}
            ilist.insert(ih, ifact.createLoad(type[argVal.num-1],
                    lvgen[argVal.num-1+offset].getIndex()));
            ilist.insert(ih, ifact.createInvoke(type[argVal.num-1].toString(),
                    "toString", Type.STRING, new Type[] {}, kind));
            // Call the pattern check
            ilist.insert(ih, ifact.createInvoke(
                    "java.util.regex.Pattern", "matches",
                    Type.BOOLEAN, new Type[]
                    {Type.STRING, Type.getType(CharSequence.class)},
                    Constants.INVOKESTATIC));
            if (iBlock.bopContainer != BopContainer.not)
            {
                ilist.insert(ih,
                    ifact.createBranchInstruction(
                        Constants.IFEQ, falseIh)); // false == 0
            }
            else
            {
                ilist.insert(ih,
                    ifact.createBranchInstruction(
                        Constants.IFNE, falseIh)); // true > 0
            }
        }
        else if (argVal.valExpr instanceof TrueExpr)
        {
            // We just care if it's null or not
            ilist.insert(ih, ifact.createLoad(type[argVal.num-1],
                    lvgen[argVal.num-1+offset].getIndex()));
            if (iBlock.bopContainer != BopContainer.not)
            {
                ilist.insert(ih,
                    ifact.createBranchInstruction(
                        Constants.IFNULL, falseIh)); // false == 0
            }
            else
            {
                ilist.insert(ih,
                    ifact.createBranchInstruction(
                        Constants.IFNONNULL, falseIh)); // true > 0
            }
        }
        
        // Return the list of instructions
        pData.returnIlist = ilist;
        return;
        
    }

    private static void buildArgValCheckInvoke(
            ArgVal argVal,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] type)
    {
        
        // Locals
        LocalVariableGen[] lvgen;
        InstructionFactory ifact = new InstructionFactory(pData.cgen);
        
        // Initialize the new InstructionList
        InstructionList ilist = iBlock.ilist;
        
        // Since we might have any kinds of types and any number of values
        // to dig through on the stack, we need to create local container
        // variables for each
        lvgen = new LocalVariableGen[type.length];
        for (int i = type.length-1; i >= (argVal.num-1); i--)
        {
            lvgen[i] = pData.mgen.addLocalVariable(
                UniqueNameGen.findUniqueLocalVariableName(pData.mgen, "lv"), type[i], null, null);
        }
        pData.mgen.stripAttributes(true);
        
        // False branches eventually jump to the nop at the end of the list
        InstructionHandle falseIh = ilist.getEnd();
        InstructionHandle ih = falseIh;
        
        // If we're within an OR, we need to insert a new nop BEFORE the
        // false position to act as the truth branch (if there isn't one
        // there already)
        InstructionHandle trueIh = ih;
        if (iBlock.bopContainer == BopContainer.or)
        {
            if (ilist.getLength() >= 2)
            {
                if (!(falseIh.getPrev().getInstruction() instanceof NOP))
                    trueIh = ilist.insert(falseIh, ifact.NOP);
                else
                    trueIh = falseIh.getPrev();
            }
            else
            {
                trueIh = ilist.insert(falseIh, ifact.NOP);
            }
            // Realign our insertion point to precede the true branch
            ih = trueIh;
            // Also, if the last non-nop instruction was a branch pointing to
            // the false branch, we need to flip it and redirect the jump to the
            // truth branch -- thus a false implies that we fall through to the
            // new check that we are currently inserting
            if (ilist.getLength() >= 3)
            {
                InstructionHandle tempIh = trueIh.getPrev();
                if (tempIh.getInstruction() instanceof IfInstruction)
                {
                    IfInstruction falseBranch =
                            (IfInstruction)tempIh.getInstruction();
                    if (falseBranch.containsTarget(falseIh))
                    {
                        falseBranch = falseBranch.negate();
                        falseBranch.setTarget(trueIh);
                        tempIh.setInstruction(falseBranch);
                    }
                }
            }
        }
        
        // Now, we pull off everything until we reach the index we care about
        for (int i = type.length-1; i >= (argVal.num-1); i--)
            ilist.insert(ih, ifact.createStore(type[i], lvgen[i].getIndex()));
        // We have the value stored, so restore the stack contents
        for (int i = argVal.num-1; i < type.length; i++)
            ilist.insert(ih, ifact.createLoad(type[i], lvgen[i].getIndex()));
        if (argVal.valExpr instanceof IntCmp)
        {
            // Now push the value for comparison
            ilist.insert(ih, ifact.createLoad(type[argVal.num-1],
                    lvgen[argVal.num-1].getIndex()));
            ilist.insert(ih, new PUSH(pData.cpgen, ((IntCmp)argVal.valExpr).n));
            // The if is dynamic to the type of the expression in the policy
            if (argVal.valExpr instanceof IntNE)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPEQ, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPNE, falseIh));
                }
            }
            else if (argVal.valExpr instanceof IntLT)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPGE, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPLT, falseIh));
                }
            }
            else if (argVal.valExpr instanceof IntGT)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPLE, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPGT, falseIh));
                }
            }
            else if (argVal.valExpr instanceof IntLE)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPGT, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPLE, falseIh));
                }
            }
            else if (argVal.valExpr instanceof IntGE)
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPLT, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPGE, falseIh));
                }
            }
            else // IntEQ
            {
                if (iBlock.bopContainer != BopContainer.not)
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPNE, falseIh));
                }
                else
                {
                    ilist.insert(ih,
                        ifact.createBranchInstruction(
                            Constants.IF_ICMPEQ, falseIh));
                }
            }
        }
        else if (argVal.valExpr instanceof StrEq)
        {
            // Create a regex comparison operation (Load the regular expression
            // first, then the input string, since the method takes them
            // in that order)
            ilist.insert(ih, new PUSH(pData.cpgen,
                    ((StrEq)argVal.valExpr).regex));
            // Obtain the result of toString()
            short kind = Constants.INVOKEVIRTUAL;
            try {
                JavaClass jclass = Repository.lookupClass(type[argVal.num-1].toString());
                if (jclass.isInterface())
                    kind = Constants.INVOKEINTERFACE;
            } catch (ClassNotFoundException ex) {/*Ignore for now*/}
            ilist.insert(ih, ifact.createLoad(type[argVal.num-1],
                    lvgen[argVal.num-1].getIndex()));
            ilist.insert(ih, ifact.createInvoke(type[argVal.num-1].toString(),
                    "toString", Type.STRING, new Type[] {}, kind));
            // Call the pattern check
            ilist.insert(ih, ifact.createInvoke(
                    "java.util.regex.Pattern", "matches",
                    Type.BOOLEAN, new Type[]
                    {Type.STRING, Type.getType(CharSequence.class)},
                    Constants.INVOKESTATIC));
            if (iBlock.bopContainer != BopContainer.not)
            {
                ilist.insert(ih,
                    ifact.createBranchInstruction(
                        Constants.IFEQ, falseIh)); // false == 0
            }
            else
            {
                ilist.insert(ih,
                    ifact.createBranchInstruction(
                        Constants.IFNE, falseIh)); // true > 0
            }
        }
        else if (argVal.valExpr instanceof TrueExpr)
        {
            // We just care if it's null or not
            ilist.insert(ih, ifact.createLoad(type[argVal.num-1],
                    lvgen[argVal.num-1].getIndex()));
            if (iBlock.bopContainer != BopContainer.not)
            {
                ilist.insert(ih,
                    ifact.createBranchInstruction(
                        Constants.IFNULL, falseIh)); // false == 0
            }
            else
            {
                ilist.insert(ih,
                    ifact.createBranchInstruction(
                        Constants.IFNONNULL, falseIh)); // true > 0
            }
        }
        
        // Return the list of instructions
        pData.returnIlist = ilist;
        return;
        
    }

    private static void buildArgValCheckGetField(
            ArgVal argVal,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] type)
    {
        // This method should never be called
    }

    private static void buildArgValCheckPutField(
            ArgVal argVal,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] type)
    {
        // Functionally equivalent to the invoke version, only the type should
        // be of one parameter
        buildArgValCheckInvoke(argVal, pData, iBlock, type);
    }

    static void buildTargetCheck(
            Target target,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] argType)
    {
        // Handle differently depending on what sort of instruction
        // we're working with
        if (pData.instruction == null)
            buildTargetCheckExecution(target, pData, iBlock, argType);
        else if (pData.instruction instanceof InvokeInstruction)
            buildTargetCheckInvoke(target, pData, iBlock, argType);
        else if (pData.instruction instanceof FieldInstruction &&
                 pData.instruction.getName().startsWith("get"))
            buildTargetCheckGetField(target, pData, iBlock, argType);
        else if (pData.instruction instanceof FieldInstruction)
            buildTargetCheckPutField(target, pData, iBlock, argType);
    }
    
    private static void buildTargetCheckInvoke(
            Target target,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] argType)
    {
        
        // Locals
        LocalVariableGen[] lvgen;
        InstructionFactory ifact = new InstructionFactory(pData.cgen);
        InvokeInstruction invoke = (InvokeInstruction)pData.instruction;
        
        // Initialize the new InstructionList
        InstructionList ilist = iBlock.ilist;
        
        // Since we might have any kinds of types and any number of values
        // to dig through on the stack, we need to create local container
        // variables for each
        lvgen = new LocalVariableGen[argType.length];
        for (int i = argType.length-1; i >= 0; i--)
        {
            lvgen[i] = pData.mgen.addLocalVariable(
                UniqueNameGen.findUniqueLocalVariableName(pData.mgen, "lv"), argType[i], null, null);
        }
        pData.mgen.stripAttributes(true);
        
        // False branches eventually jump to the nop at the end of the list
        InstructionHandle falseIh = ilist.getEnd();
        InstructionHandle ih = falseIh;
        
        // If we're within an OR, we need to insert a new nop BEFORE the
        // false position to act as the truth branch (if there isn't one
        // there already)
        InstructionHandle trueIh = ih;
        if (iBlock.bopContainer == BopContainer.or)
        {
            if (ilist.getLength() >= 2)
            {
                if (!(falseIh.getPrev().getInstruction() instanceof NOP))
                    trueIh = ilist.insert(falseIh, ifact.NOP);
                else
                    trueIh = falseIh.getPrev();
            }
            else
            {
                trueIh = ilist.insert(falseIh, ifact.NOP);
            }
            // Realign our insertion point to precede the true branch
            ih = trueIh;
            // Also, if the last non-nop instruction was a branch pointing to
            // the false branch, we need to flip it and redirect the jump to the
            // truth branch -- thus a false implies that we fall through to the
            // new check that we are currently inserting
            if (ilist.getLength() >= 3)
            {
                InstructionHandle tempIh = trueIh.getPrev();
                if (tempIh.getInstruction() instanceof IfInstruction)
                {
                    IfInstruction falseBranch =
                            (IfInstruction)tempIh.getInstruction();
                    if (falseBranch.containsTarget(falseIh))
                    {
                        falseBranch = falseBranch.negate();
                        falseBranch.setTarget(trueIh);
                        tempIh.setInstruction(falseBranch);
                    }
                }
            }
        }
        
        // Now, we pull off everything until we reach the index we care about
        for (int i = argType.length-1; i >= 0; i--)
            ilist.insert(ih, ifact.createStore(argType[i], lvgen[i].getIndex()));
        // We have the value stored, so restore the stack contents
        for (int i = 0; i < argType.length; i++)
            ilist.insert(ih, ifact.createLoad(argType[i], lvgen[i].getIndex()));
        
        // Do the instanceof check
        ilist.insert(ih, ifact.createLoad(argType[0], lvgen[0].getIndex()));
        ilist.insert(ih, ifact.createInstanceOf((ReferenceType)Misc.constructType(target.typeName)));
        if (iBlock.bopContainer != BopContainer.not)
        {
            ilist.insert(ih, ifact.createBranchInstruction(
                    Constants.IFEQ, falseIh));
        }
        else
        {
            ilist.insert(ih, ifact.createBranchInstruction(
                    Constants.IFNE, falseIh));
        }
        
        // Return the list of instructions
        pData.returnIlist = ilist;
        return;
        
    }

    private static void buildTargetCheckExecution(
            Target target,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] argVal)
    {
    }

    private static void buildTargetCheckGetField(
            Target target,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] argVal)
    {
        // Functionally equivalent to the invoke version
        buildTargetCheckInvoke(target, pData, iBlock, argVal);
    }

    private static void buildTargetCheckPutField(
            Target target,
            PointcutData pData,
            InstructionBlock iBlock,
            Type[] argVal)
    {
        // Functionally equivalent to the invoke version
        buildTargetCheckInvoke(target, pData, iBlock, argVal);
    }
}
