/*
 * Handles math compilation operations.
 */

package rewrite.rewriter;

import java.util.*;

import rewrite.rewriter.*;
import rewrite.util.*;

import org.apache.bcel.generic.*;
import org.apache.bcel.Constants;

import org.nfunk.jep.*;
import org.lsmp.djep.xjep.*;
import org.nfunk.jep.function.*;

import spox.policy.*;
import spox.policy.PolicyComponent.*;

/**
 * Provides methods for compiling bytecode instructions for math expressions.
 * Meant primarily for use with 'nodes' tags.
 */
public class ExpressionCodeBuilder
{

    private static TreeUtils treeUtils = new TreeUtils();
    private static String varName = "i";
    
    public static InstructionList buildFromCheck(
            Nodes nodes,
            Edge edge,
            HashMap<String, Forall> forallVar,
            HashMap<String, Integer> computedForall,
            int lvIndex,
            ClassGen cgen,
            MethodGen mgen)
        throws UnenforceableException
    {
        JEP jep = new JEP();
        InstructionList ilist = new InstructionList();
        InstructionFactory ifact = new InstructionFactory(cgen);
        double rand = Math.random();
        ConstantPoolGen cpgen = cgen.getConstantPool();
        
        Iterator<String> iter = forallVar.keySet().iterator();
        while (iter.hasNext())
            jep.addVariable(iter.next(), 0);
        jep.addVariable("__" + rand + " state__var__",0);
        
        try
        {
            jep.setImplicitMul(true);
            Node topNode = jep.parse(nodes.from.getString());
            //outputTree(topNode);
            // Given, for example, x=i+4, solve for i in terms of x
            // For now, we assume only one variable on the righthand side
            Node newNode = solveForVar(topNode);
            ASTVarNode varNode = new ASTVarNode(0);
            varNode.setVar(jep.getVar("__" + rand + " state__var__"));
            if (newNode == null)
                newNode = varNode;
            else
            {
                newNode.jjtAddChild(varNode, 0);
                varNode.jjtSetParent(newNode);
                newNode.jjtClose();
            }
            
            Forall forall = forallVar.get(varName);
            
            Node newTopNode = getTopParentNode(newNode);

            outputTree(newTopNode);
//            System.out.println(jep.evaluate(newTopNode));
            
            // Build the math computation and the check
            HashMap<String, Integer> varLvIndex = new HashMap<String, Integer>();
            varLvIndex.put(varNode.getName(), lvIndex);
            ilist = buildMathIlist(newTopNode, varLvIndex, cgen);
            ilist.append(ifact.D2I);
            if (computedForall.containsKey(varName))
            {
                InstructionHandle ih =
                        ilist.append(ifact.NOP);
                // The value of i was computed before, so check it against that
                ilist.insert(ih, ifact.createLoad(Type.INT,
                        computedForall.get(varName)));
                ilist.insert(ih, ifact.createBranchInstruction(
                        Constants.IF_ICMPNE, ih));
            }
            else
            {
                ilist.append(ifact.DUP);
                InstructionHandle ih =
                        ilist.append(ifact.POP);
                // Check the result against the boundaries of the forall
                ilist.insert(ih, new PUSH(cpgen,
                        forall.from));
                ilist.insert(ih,
                        ifact.createBranchInstruction(
                        Constants.IF_ICMPLT, ih));
                ilist.insert(ih, ifact.DUP);
                ilist.insert(ih, new PUSH(cpgen,
                        forall.to));
                ilist.insert(ih,
                        ifact.createBranchInstruction(
                        Constants.IF_ICMPGT, ih));
                // Create a local variable to store the result in
                LocalVariableGen lvgen = mgen.addLocalVariable(
                        UniqueNameGen.findUniqueLocalVariableName(mgen, varName),
                        Type.INT, null, null);
                computedForall.put(varName, lvgen.getIndex());
                ilist.insert(ih, ifact.createStore(Type.INT, lvgen.getIndex()));
            }

        }
        catch (ParseException ex)
        {
            throw new UnenforceableException(
                    "Expression '" + nodes.from.getString() +
                    "' could not be parsed:\n" + ex.getMessage());
        }
        
        return ilist;
    }
    
    public static InstructionList buildTo(
            Nodes nodes,
            Edge edge,
            HashMap<String, Integer> computedForall,
            ClassGen cgen)
        throws UnenforceableException
    {
        JEP jep = new JEP();
        InstructionList ilist = new InstructionList();
        InstructionFactory ifact = new InstructionFactory(cgen);
        ConstantPoolGen cpgen = cgen.getConstantPool();
        
        Iterator<String> iter = computedForall.keySet().iterator();
        while (iter.hasNext())
            jep.addVariable(iter.next(), 0);
        
        try
        {
            Node topNode = jep.parse(nodes.to.getString());
            ilist = buildMathIlist(topNode, computedForall, cgen);
            ilist.append(ifact.D2I);
        }
        catch (ParseException ex)
        {
            throw new UnenforceableException(
                    "Expression '" + nodes.to.getString() +
                    "' could not be parsed:\n" + ex.getMessage());
        }
        
        return ilist;
    }

    /**
     * Recursively scans through a JEP tree and builds the instructions to
     * compute the result.
     * @param topNode
     * @return
     */
    private static InstructionList buildMathIlist(
            Node topNode,
            HashMap<String, Integer> varLvIndex,
            ClassGen cgen)
    {
        InstructionList ilist = new InstructionList();
        InstructionFactory ifact = new InstructionFactory(cgen);
        Node[] child = TreeUtils.getChildrenAsArray(topNode);
        
        if (treeUtils.isOperator(topNode))
        {
            String operator = treeUtils.getOperator(topNode).getName();
            if (operator.equals("UMinus") &&
                    topNode.jjtGetNumChildren() == 1)
            {
                ilist.append(buildMathIlist(child[0], varLvIndex, cgen));
                ilist.append(ifact.DNEG);
            }
            else if (operator.equals("+") &&
                topNode.jjtGetNumChildren() == 2)
            {
                ilist.append(buildMathIlist(child[0], varLvIndex, cgen));
                ilist.append(buildMathIlist(child[1], varLvIndex, cgen));
                ilist.append(ifact.DADD);
            }
            else if (operator.equals("-") &&
                topNode.jjtGetNumChildren() == 2)
            {
                ilist.append(buildMathIlist(child[0], varLvIndex, cgen));
                ilist.append(buildMathIlist(child[1], varLvIndex, cgen));
                ilist.append(ifact.DSUB);
            }
            else if (operator.equals("*") &&
                topNode.jjtGetNumChildren() == 2)
            {
                ilist.append(buildMathIlist(child[0], varLvIndex, cgen));
                ilist.append(buildMathIlist(child[1], varLvIndex, cgen));
                ilist.append(ifact.DMUL);
            }
            else if (operator.equals("/") &&
                topNode.jjtGetNumChildren() == 2)
            {
                ilist.append(buildMathIlist(child[0], varLvIndex, cgen));
                ilist.append(buildMathIlist(child[1], varLvIndex, cgen));
                ilist.append(ifact.DDIV);
            }
            else if (operator.equals("^") &&
                    topNode.jjtGetNumChildren() == 2)
            {
                ilist.append(buildMathIlist(child[0], varLvIndex, cgen));
                ilist.append(buildMathIlist(child[1], varLvIndex, cgen));
                ilist.append(ifact.createInvoke("java.lang.Math", "pow",
                        Type.DOUBLE, new Type[] {Type.DOUBLE, Type.DOUBLE},
                        Constants.INVOKESTATIC));
            }
            else if (operator.equals("ln") &&
                    topNode.jjtGetNumChildren() == 1)
            {
                ilist.append(buildMathIlist(child[0], varLvIndex, cgen));
                ilist.append(ifact.createInvoke("java.lang.Math", "log",
                        Type.DOUBLE, new Type[] {Type.DOUBLE},
                        Constants.INVOKESTATIC));
            }
        }
        else if (treeUtils.isConstant(topNode))
        {
            ilist.append(ifact.createConstant(((ASTConstant)topNode).getValue()));
        }
        else if (treeUtils.isVariable(topNode))
        {
            ilist.append(ifact.createLoad(Type.INT, varLvIndex.get(
                    ((ASTVarNode)topNode).getVar().getName())));
            ilist.append(ifact.I2D);
        }
        
        return ilist;
    }
    
    private static void outputTree(Node topNode)
    {
        System.out.println(topNode);
        if (topNode != null)
        {
        Node[] child = TreeUtils.getChildrenAsArray(topNode);
        for (int i = 0; i < child.length; i++)
            outputTree(child[i]);
        }
    }
    
    private static Node getTopParentNode(Node node)
    {
        if (node.jjtGetParent() != null)
            return getTopParentNode(node.jjtGetParent());
        else
            return node;
    }
    
    private static Node solveForVar(Node topNode)
            throws UnenforceableException
    {
        ASTFunNode retNode = null;
        
        int numVars = countVars(topNode);
        if (numVars < 1)
            throw new UnenforceableException(
                    "Nodes 'from' expression contains no iteration variable.");
        else if (numVars > 1)
            throw new UnenforceableException(
                    "Nodes 'from' expression contains more than one iteration variable.");

        Node[] child = TreeUtils.getChildrenAsArray(topNode);
        
        if (treeUtils.isOperator(topNode))
        {
            String operator = treeUtils.getOperator(topNode).getName();
            if (operator.equals("UMinus"))
            {
                if (countVars(child[0]) > 0)
                {
                    retNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("UMinus","UMinus",new UMinus()));
                    retNode.jjtOpen();
                    Node parentNode = solveForVar(child[0]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(retNode, 0);
                        retNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                }
            }
            else if (operator.equals("+") &&
                topNode.jjtGetNumChildren() == 2)
            {
                if (countVars(child[0]) == 0)
                {
                    retNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("-","-",new Subtract()));
                    retNode.jjtOpen();
                    retNode.jjtAddChild(child[0], 1);
                    Node parentNode = solveForVar(child[1]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(retNode, 0);
                        retNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                }
                else
                {
                    // Var must be in left-hand side
                    retNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("-","-",new Subtract()));
                    retNode.jjtOpen();
                    retNode.jjtAddChild(child[1], 1);
                    Node parentNode = solveForVar(child[0]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(retNode, 0);
                        retNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                }
            }
            if (operator.equals("-") &&
                topNode.jjtGetNumChildren() == 2)
            {
                if (countVars(child[0]) == 0)
                {
                    // Subtract left-hand part, then negate by multiplying by -1
                    retNode = new ASTFunNode(0);
                    ASTFunNode negNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("-","-",new Subtract()));
                    retNode.jjtOpen();
                    retNode.jjtAddChild(child[0], 1);
                    negNode.setOperator(new Operator("*","*",new Multiply()));
                    negNode.jjtOpen();
                    ASTConstant minusOneNode = new ASTConstant(0);
                    minusOneNode.setValue(-1.0);
                    negNode.jjtAddChild(minusOneNode, 1);
                    Node parentNode = solveForVar(child[1]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(negNode, 0);
                        negNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                    negNode.jjtAddChild(retNode, 0);
                    retNode.jjtSetParent(negNode);
                    negNode.jjtClose();
                }
                else
                {
                    // Var must be in left-hand side
                    retNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("+","+",new Add()));
                    retNode.jjtOpen();
                    retNode.jjtAddChild(child[1], 1);
                    Node parentNode = solveForVar(child[0]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(retNode, 0);
                        retNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                }
            }
            else if (operator.equals("*") &&
                topNode.jjtGetNumChildren() == 2)
            {
                if (countVars(child[0]) == 0)
                {
                    retNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("/","/",new Divide()));
                    retNode.jjtOpen();
                    retNode.jjtAddChild(child[0], 1);
                    Node parentNode = solveForVar(child[1]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(retNode, 0);
                        retNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                }
                else
                {
                    // Var must be in left-hand side
                    retNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("/","/",new Divide()));
                    retNode.jjtOpen();
                    retNode.jjtAddChild(child[1], 1);
                    Node parentNode = solveForVar(child[0]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(retNode, 0);
                        retNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                }
            }
            if (operator.equals("/") &&
                topNode.jjtGetNumChildren() == 2)
            {
                if (countVars(child[0]) == 0)
                {
                    // Divide left-hand part, then flip the fraction by taking
                    // a 1 and dividing that by the part (i.e., 1/x)
                    retNode = new ASTFunNode(0);
                    ASTFunNode flipNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("/","/",new Divide()));
                    retNode.jjtOpen();
                    retNode.jjtAddChild(child[0], 1);
                    flipNode.setOperator(new Operator("/","/",new Divide()));
                    flipNode.jjtOpen();
                    ASTConstant oneNode = new ASTConstant(0);
                    oneNode.setValue(1.0);
                    flipNode.jjtAddChild(oneNode, 0);
                    Node parentNode = solveForVar(child[1]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(flipNode, 0);
                        flipNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                    flipNode.jjtAddChild(retNode, 1);
                    retNode.jjtSetParent(flipNode);
                    flipNode.jjtClose();
                }
                else
                {
                    // Var must be in left-hand side
                    retNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("*","*",new Multiply()));
                    retNode.jjtOpen();
                    retNode.jjtAddChild(child[1], 1);
                    Node parentNode = solveForVar(child[0]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(retNode, 0);
                        retNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                }
            }
            if (operator.equals("^") &&
                topNode.jjtGetNumChildren() == 2)
            {
                if (countVars(child[0]) == 0)
                {
                    // Case for x=2^i -> i=log2(x)
                    // Note that log2(x) = ln(x)/ln(2)
                    // So we take the ln of the left side and divide that
                    // by the ln of the base.
                    retNode = new ASTFunNode(0);
                    ASTFunNode divNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("ln","ln",new NaturalLogarithm()));
                    retNode.jjtOpen();
                    divNode.setOperator(new Operator("/","/",new Divide()));
                    divNode.jjtOpen();
                    ASTFunNode lnNode = new ASTFunNode(0);
                    lnNode.setOperator(new Operator("ln","ln",new NaturalLogarithm()));
                    lnNode.jjtOpen();
                    lnNode.jjtAddChild(child[0], 0);
                    lnNode.jjtClose();
                    divNode.jjtAddChild(lnNode, 1);
                    Node parentNode = solveForVar(child[1]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(divNode, 0);
                        divNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                    divNode.jjtAddChild(retNode, 0);
                    retNode.jjtSetParent(divNode);
                    divNode.jjtClose();
                }
                else
                {
                    // Case for x=i^2 -> i=x^(1/2)
                    // Take the right-hand side, flip it by doing 1/a,
                    // then treat that as an exponent of x.
                    retNode = new ASTFunNode(0);
                    retNode.setOperator(new Operator("^","^",new Power()));
                    retNode.jjtOpen();
                    ASTFunNode flipNode = new ASTFunNode(0);
                    flipNode.setOperator(new Operator("/","/",new Divide()));
                    flipNode.jjtOpen();
                    ASTConstant oneNode = new ASTConstant(0);
                    oneNode.setValue(1.0);
                    flipNode.jjtAddChild(oneNode, 0);
                    flipNode.jjtAddChild(child[1], 1);
                    flipNode.jjtClose();
                    retNode.jjtAddChild(flipNode, 1);
                    flipNode.jjtSetParent(retNode);
                    Node parentNode = solveForVar(child[0]);
                    if (parentNode != null)
                    {
                        parentNode.jjtAddChild(retNode, 0);
                        retNode.jjtSetParent(parentNode);
                        parentNode.jjtClose();
                    }
                }
            }
        }
        
        return retNode;
    }

    private static int countVars(Node node)
    {
        int retVal = 0;
        
        if (treeUtils.isVariable(node))
        {
            retVal++;
            varName = ((ASTVarNode)node).getName();
        }
        
        Node[] child = TreeUtils.getChildrenAsArray(node);
        for (int i = 0; i < child.length; i++)
            retVal = retVal + countVars(child[i]);
        
        return retVal;
    }
    
    public static void parseEquation(String expression)
            throws UnenforceableException
    {

        TreeUtils treeUtils = new TreeUtils();
        
        JEP jep = new JEP();
        jep.setAllowUndeclared(true);
        jep.parseExpression(expression);
        
        Node[] nodeArray = treeUtils.getChildrenAsArray(jep.getTopNode());
        System.out.println(treeUtils.getOperator(nodeArray[0].jjtGetParent()).getName());
        for (int i = 0; i < nodeArray.length; i++)
        {
            if (treeUtils.isVariable(nodeArray[i]))
                System.out.println(treeUtils.getName(nodeArray[i]));
            else if (treeUtils.isOperator(nodeArray[i]))
                System.out.println(treeUtils.getOperator(nodeArray[i]).toString());
            else if (treeUtils.isInteger(nodeArray[i]))
                System.out.println(treeUtils.getValue(nodeArray[i]));
        }

        if (jep.getErrorInfo() != null)
            throw new UnenforceableException("Expression '" + expression +
                    "' could not be parsed:\n" + jep.getErrorInfo());
    }
}
