/*
 * Handles rewriting of JAR files according to a given security policy
 */

package rewrite.rewriter;

import rewrite.util.*;

import java.io.*;
import java.util.*;
import java.util.jar.*;

import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.Repository;
import org.apache.bcel.util.*;
import org.apache.bcel.Constants;

import spox.policy.*;
import spox.policy.pointcut.*;
import spox.policy.PolicyComponent.*;

public class JarRewriter
{
    
    // Constants
    public static final String FILE_SEPARATOR = System.getProperty("file.separator");
    public static final String PATH_SEPARATOR = System.getProperty("path.separator");
    public static final String LINE_SEPARATOR = System.getProperty("line.separator");
    
    // Class fields
    private File file;
    private JarFile jarfile;
    private String outFilename;
    private Policy policy;
    private HashMap<String, ClassGen> classMap = new HashMap<String, ClassGen>();
    
    // Constructor (requires file name arguments!)
    public JarRewriter
            (String jarFilename,
            Policy policy,
            String outFilename)
        throws Exception
    {
        file = new File(jarFilename);
        jarfile = new JarFile(file);
        this.outFilename = outFilename;
        this.policy = policy;
    }
    
    // Methods
    public void rewriteJar()
        throws Exception
    {
        // Systematically rewrites the given JAR file's underlying .class files.
        // First, a class security/policy/Policy is added (Unless it already
        // exists, in which case it's given another name that isn't taken).
        // Next, each individual class file is analyzed and rewritten with
        // checks.
        
        // Locals
        PolicyClass policyClass;
        
        FileOutputStream dest;
        JarOutputStream outjar = null;
        Enumeration entries;
        JarEntry entry;
        InputStream is;
        
        JavaClass jclass;
        ClassGen cgen;
        String policyClassName = "";
        HashMap<String, StateHolder> wrapperClassMap = new HashMap<String, StateHolder>();
        
        ClassRewriter classRewriter;
        
        try
        {
            // Set the classpath to the JAR we're rewriting
            Repository.clearCache();
            String manifestClassPath = jarfile.getManifest().getMainAttributes().getValue("Class-Path");
            if (manifestClassPath == null)
                manifestClassPath = "";
            else
                manifestClassPath = manifestClassPath.trim().
                            replace(".jar ", ".jar" + PATH_SEPARATOR + file.getParent() + FILE_SEPARATOR).
                            replace("/", FILE_SEPARATOR);
            Repository.setRepository(SyntheticRepository.getInstance(
                    new ClassPath(
                        (jarfile.getName() + " " + manifestClassPath))));
            System.out.println("Class path:\n" +
                    Repository.getRepository().getClassPath().toString());
            
            // Type check the policy before proceeding
            PolicyTypeCheck.checkPolicy(policy);
            
            // Create a new file to output the updated JAR
            dest = new FileOutputStream(outFilename);
            outjar = new JarOutputStream(new BufferedOutputStream(dest));

//            wrapperClassMap = buildWrappers(outjar);
            // Make an array of cgen classname references so that we don't
            // overwrite anything we just did
//            ArrayList<String> wrapClassFileName = new ArrayList<String>();
//            Iterator<InstanceStateHolder> iter = wrapperClassMap.values().iterator();
//            while (iter.hasNext())
//                wrapClassFileName.add(iter.next().
//                        getClassName().replace('.', '/') + ".class");
            
            // Create a new class which will eventually contain the central policy code
            ClassGen policyCgen = new ClassGen(
                    "security.policy.Policy",
                    "java.lang.Object",
                    null,
                    Constants.ACC_PUBLIC, null);
            policyClassName = policyCgen.getClassName();

            // Add this information to the manifest for reference purposes
            // (We'll write it in later)
            Manifest manifest = jarfile.getManifest();
            manifest.getMainAttributes().putValue(
                    UniqueNameGen.findUniqueManifestAttributeName(manifest, "Security-Policy"),
                    policyClassName);

            // Start scanning through the entries for other classes
            entries = jarfile.entries();

            while (entries.hasMoreElements())
            {
                entry = (JarEntry)entries.nextElement();

                System.out.println("Extracting file: " + entry.getName());

                is = jarfile.getInputStream(entry);

                // If the entry is a class file, add it to our repository
                // Otherwise just copy as is (except for the manifest)
                if (entry.getName().endsWith(".class"))
                {

                    jclass = new ClassParser(
                                is, entry.getName()).parse();
                    cgen = new ClassGen(jclass);
                    classMap.put(cgen.getClassName(), cgen);

                }
                else if (entry.getName().equals(jarfile.MANIFEST_NAME))
                {
                    System.out.println("Writing " + entry.getName() + "...");
                    outjar.putNextEntry(new JarEntry(entry.getName()));
                    manifest.write(outjar);
                    System.out.println("Manifest updated");
                    outjar.closeEntry();
                }
                else
                {
                    System.out.println("Writing " + entry.getName() + "...");
                    outjar.putNextEntry(entry);
                    int b = is.read();
                    while (b != -1)
                    {
                        outjar.write(b);
                        b = is.read();
                    }
                    outjar.closeEntry();
                }

                is.close();

            }

            // Handle instance states
            insertInstanceStates();
            
            // Create the policy class
            policyClass = new PolicyClass(policyCgen, policy);

            // Make sure the new policy class doesn't violate the policy!
            classRewriter = new ClassRewriter(
                    policyClass, policyCgen, policy, classMap);
            if (classRewriter.rewriteClass())
                throw new UnenforceableException("Cannot enforce policy on injected Policy class.");

            // Rewrite the classes
            Iterator<ClassGen> iterator = classMap.values().iterator();
            while (iterator.hasNext())
            {
                cgen = iterator.next();
                System.out.println("Rewriting " + cgen.getClassName());
                classRewriter = new ClassRewriter(
                       policyClass, cgen, policy, classMap);
                classRewriter.rewriteClass();
                cgen = classRewriter.getCgen();
                classMap.put(cgen.getClassName(), cgen);
            }
            
            // Finalize the wrappers
            StateHolder.finalizeStatefuls();

            // Add all of the classes to the JAR
            iterator = classMap.values().iterator();
            while (iterator.hasNext())
            {
                cgen = iterator.next();
                System.out.println("Writing " + cgen.getClassName().replace('.','/') + ".class" + "...");
                outjar.putNextEntry(new JarEntry(cgen.getClassName().replace('.','/') + ".class"));
                cgen.getJavaClass().dump(outjar);
            }

            // Inject the policy class
            System.out.println("Writing " + policyCgen.getClassName().replace('.','/') + ".class" + "...");
            outjar.putNextEntry(
                    new JarEntry(policyCgen.getClassName().replace('.','/') + ".class"));
            policyCgen.getJavaClass().dump(outjar);

            System.out.println("New class " + policyClassName + " inserted.");

            outjar.close();
            jarfile.close();

            System.out.println("Finished writing to " + outFilename + ".");
        
        }
        catch (Exception ex)
        {
            jarfile.close();
            throw ex;
        }

    }
    
    public void insertInstanceStates()
            throws UnenforceableException, ClassNotFoundException
    {
        // Inject instance state variables into all relevant classes
        // For classes which do not exist in the JAR, build wrappers
        
        // Initialize the instance state holder data
        StateHolder.initialize();
        StateHolder.setPolicy(policy);
        StateHolder.setClassMap(classMap);
        
        // Get the states
        ArrayList<State> state = policy.getData().state;
        
        // Find instance state variables
        for (int i = 0; i < state.size(); i++)
        {
            if (!state.get(i).objName.isEmpty())
            {
                StateHolder.injectState(state.get(i).objName, state.get(i));
            }
        }
        
    }
    
//    public HashMap<String, InstanceStateHolder> buildWrappers(JarOutputStream outjar)
//            throws Exception
//    {
//        // Constructs wrapper classes for each instance state parameter in
//        // the policy, IF the original class is not contained in the JAR file
//        // Return value is a HashMap mapping original class names to the
//        // corresponding wrapper class names
//        
//        InstanceStateHolder wrapperClass;
//        String superclassName = "";
//        HashMap<String, InstanceStateHolder> wrapperMap = new HashMap<String, InstanceStateHolder>();
//        
//        // Get all states in the policy
//        ArrayList<State> state = policy.getData().state;
//        
//        for (int i = 0; i < state.size(); i++)
//        {
//            if (!state.get(i).objName.isEmpty())
//            {
//                wrapperMap.putAll(buildWrapper(
//                        state.get(i).objName, state.get(i), wrapperMap));
//            }
//        }
//        
//        // Make sure our new classes don't violate the policy
//        // (We're being tricky here, since the policy class hasn't
//        // been generated yet; if that policy class is referred to
//        // at all by the rewriter, a NullPointerException will be thrown.
//        // We catch that exception and throw an UnenforceableException.)
//        Iterator<InstanceStateHolder> iter = wrapperMap.values().iterator();
//        while (iter.hasNext())
//        {
//            ClassGen cgen = iter.next().getCgen();
//            try
//            {
//                ClassRewriter classRewriter = new ClassRewriter(
//                        null, cgen, policy, null);
//            }
//            catch (Exception ex)
//            {
//                throw new UnenforceableException("Cannot enforce policy on injected wrapper class.");
//            }
//            if (!wrapperMap.containsKey(cgen.getClassName()))
//            {
//                // If this wasn't a pre-existing class, add it to the JAR
//                outjar.putNextEntry(
//                        new JarEntry(cgen.getClassName().replace('.','/') + ".class"));
//                cgen.getJavaClass().dump(outjar);
//            }
//        }
//
//        return wrapperMap;
//        
//    }
    
//    public HashMap<String, InstanceStateHolder> buildWrapper(
//            String objName,
//            State state,
//            HashMap<String,InstanceStateHolder> wrapperMap)
//            throws Exception
//    {
//        // Wraps based upon a particular state
//        // Recursively generates wrappers for all superclasses
//        // up to java.lang.Object
//
//        // Locals
//        InstanceStateHolder wrapperClass;
//        String className = "";
//        String superclassName = "";
//        boolean addField;
//
//        // If the wrapper class has not yet been created, make it
//        // Otherwise, simply add this state to the pre-existing wrapper
//        if (!wrapperMap.containsKey(objName))
//        {
//            ClassGen cgen;
//            addField = true;
//            JarEntry jarEntry = jarfile.getJarEntry(
//                    objName.replace('.', '/') + ".class");
//            if (jarEntry == null)
//            {
//                JavaClass superclass = Repository.lookupClass(
//                        Class.forName(objName).getSuperclass());
//                superclassName = superclass.getClassName();
//                if (!superclassName.equals("java.lang.Object"))
//                {
//                    wrapperMap.putAll(buildWrapper(
//                            superclassName, state, wrapperMap));
//                    addField = false;
//                }
//                cgen = new ClassGen(
//                        UniqueNameGen.findUniqueClassFileName(
//                            jarfile, "security/wrapper/Wrap_" +
//                            objName.replace('.','_')).replace('/','.'),
//                        superclassName, null,
//                        Constants.ACC_PUBLIC, null);
//            }
//            else
//            {
//                cgen = new ClassGen(
//                        new ClassParser(
//                        jarfile.getInputStream(jarEntry),
//                        jarEntry.getName()).parse());
//            }
//            wrapperClass = new InstanceStateHolder(cgen, state, addField);
//            wrapperMap.put(objName, wrapperClass);
//        }
//        else
//        {
//            className = objName;
//            superclassName = Class.forName(objName).getSuperclass().getName();
//            while (!superclassName.equals("java.lang.Object") &&
//                   wrapperMap.containsKey(superclassName))
//            {
//                className = superclassName;
//                superclassName = Class.forName(className).getSuperclass().getName();
//            }
//            JavaClass javaClass = Repository.lookupClass(
//                    Class.forName(className));
//            wrapperClass = wrapperMap.get(className);
//            wrapperClass.addState(state, true);
//        }
//        
//        return wrapperMap;
//    }

//    public void rewriteJar()
//    throws Exception
//    {
//        // Locals
//        FileOutputStream dest;
//        JarOutputStream outjar;
//        Enumeration entries;
//        JarEntry entry;
//        InputStream is;
//        
//        JavaClass jclass;
//        ClassGen cgen;
//        String policyClass = "";
//        String callMethod = "";
//        
//        // Create a new file to output the updated JAR
//        dest = new FileOutputStream(outFilename);
//        outjar = new JarOutputStream(new BufferedOutputStream(dest));
//
//        // Write a new class containing the central policy code
//        cgen = new ClassGen(
//                UniqueNameGen.findUniqueClassFileName(
//                    jarfile, "security/policy/Policy").replace('/','.'),
//                "java.lang.Object", null,
//                Constants.ACC_PUBLIC, null);
//        policyClass = cgen.getClassName();
//        addPolicyClass(cgen);
//        outjar.putNextEntry(
//                new JarEntry(cgen.getClassName().replace('.','/') + ".class"));
//        cgen.getJavaClass().dump(outjar);
//        
//        System.out.println("New class " + policyClass + " inserted.");
//        
//        // Add this information to the manifest for reference purposes
//        // (We'll write it in later)
//        Manifest manifest = jarfile.getManifest();
//        manifest.getMainAttributes().putValue(
//                UniqueNameGen.findUniqueManifestAttributeName(manifest, "Security-Policy"),
//                policyClass);
//
//        // Start scanning through the entries for other classes
//        entries = jarfile.entries();
//
//        while (entries.hasMoreElements())
//        {
//            entry = (JarEntry)entries.nextElement();
//
//            System.out.println("Extracting file: " + entry.getName());
//
//            is = jarfile.getInputStream(entry);
//
//            if (entry.getName().endsWith(".class"))
//            {
//
//                // Initialize jclass and cgen for this class file
//                jclass = new ClassParser(
//                        is, entry.getName()).parse();
//                cgen = new ClassGen(jclass);
//
//                // Scan for method calls
//                addChecks(policyClass, cgen);
//                // Update the file in the new jar
//                outjar.putNextEntry(new JarEntry(entry.getName()));
//                cgen.getJavaClass().dump(outjar);
//                
//            }
//            else if (entry.getName().equals(jarfile.MANIFEST_NAME))
//            {
//                outjar.putNextEntry(new JarEntry(entry.getName()));
//                manifest.write(outjar);
//                System.out.println("Manifest updated");
//            }
//            else
//            {
//                outjar.putNextEntry(entry);
//                int b = is.read();
//                while (b != -1)
//                {
//                    outjar.write(b);
//                    b = is.read();
//                }
//            }
// 
//            outjar.closeEntry();
//            is.close();
//            
//        }
//
//        outjar.close();
//        jarfile.close();
//
//        System.out.println("Finished writing to " + outFilename + ".");
//        
//    }

}
