/*
 * Here the 'policy' class containing the static counter field and the
 * abort method is created. Instances of PolicyClass are the point of
 * reference when we need information regarding the locations of the
 * counter field, the abort method, or if we want prebuilt check instructions.
 */

package rewrite.rewriter;

import rewrite.util.*;

import java.io.*;
import java.util.*;
import java.util.jar.*;

import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;
import org.apache.bcel.util.*;
import org.apache.bcel.Constants;

import spox.policy.*;
import spox.policy.pointcut.*;
import spox.policy.PolicyComponent.*;

public class PolicyClass
{
    
    // Fields
    private HashMap<Edge,Method> edgeMap = new HashMap<Edge,Method>();
    private String className = null;
    private Policy policy = null;

    // Constructor
    public PolicyClass(
            ClassGen cgen,
            Policy policy)
        throws Exception
    {
        this.policy = policy;
        this.className = cgen.getClassName();
        addPolicyClass(cgen);
    }
    
    // Accessors
    public String getEdgeMethodName(Edge edge)
    { return edgeMap.get(edge).getName(); }
    
    public String getCommitMethodName(Edge edge)
    { return edgeMap.get(edge).getName().replaceFirst("edge_", "commit_"); }
    
    public Type[] getEdgeMethodArgumentTypes(Edge edge)
    { return edgeMap.get(edge).getArgumentTypes(); }
    
    public String getClassName()
    { return className; }

    // Methods
    private void addPolicyClass
            (ClassGen cgen)
    throws Exception
    {
        // Here we will add a new class security.policy.Policy to the given
        // JAR. For each non-instance "state" element in the policy data
        // structure, we will add a static state variable and a public static
        // method for modifying that state.
        // Note that states are Integers, not ints. This allows us to use
        // null to indicate a "bad" state.
        
        // Set up the construction tools
        InstructionFactory ifact = new InstructionFactory(cgen);
        InstructionList ilist = new InstructionList();
        InstructionList cinitIlist = new InstructionList();
        ConstantPoolGen cpgen = cgen.getConstantPool();
        String cname = cgen.getClassName();
        MethodGen mg;
        Instruction instruction;
        InstructionHandle ih;
        Type IntegerType = Type.getType(java.lang.Integer.class);
        
        String fname;  // Name of the state field
        String ofname; // Original name of the state
        String fcname; // Name of the class in which the field is contained
        String mname;  // Name of a method
        String setMethodName;
        String getMethodName;
        
        boolean isInstanceState = false;
        short invokeType = Constants.INVOKESTATIC;

        HashMap<String, String[]> stateMap =
                new HashMap<String, String[]>();
        ArrayList<State> state = policy.getData().state;
        ArrayList<Forall> forall = policy.getData().forall;
        ArrayList<Edge> edge = policy.getData().edge;
        
        // Create a new 'state' variable and modifier method for each
        // state item in the policy
        for (int i = 0; i < state.size(); i++)
        {
            isInstanceState = !state.get(i).objName.isEmpty();
            if (!isInstanceState)
            {
                // Create a new Integer field for the state (note that duplicate
                // state names -- which should not be allowed by the parser --
                // will end up generating separate field names)
                fname = UniqueNameGen.findUniqueFieldName(
                        cgen, state.get(i).name);
                ofname = fname;
                fcname = cname;
                FieldGen fgen = new FieldGen(
                        Constants.ACC_PRIVATE | Constants.ACC_STATIC,
                        IntegerType, fname, cpgen);
                // Add in a temp
                FieldGen tempFgen = new FieldGen(
                        Constants.ACC_PRIVATE | Constants.ACC_STATIC,
                        IntegerType, "temp_"+fname, cpgen);

                // Add in the new fields
                cgen.addField(fgen.getField());
                cgen.addField(tempFgen.getField());

                // Add the necessary code to the static initializer
                cinitIlist.append(ifact.createNew((ObjectType)IntegerType));
                cinitIlist.append(ifact.DUP);
                // Initialize to the correct state
                cinitIlist.append(ifact.createConstant(state.get(i).start));
                cinitIlist.append(ifact.createInvoke("java.lang.Integer",
                        "<init>", Type.VOID, new Type[] {Type.INT},
                        Constants.INVOKESPECIAL));
                cinitIlist.append(ifact.DUP);
                cinitIlist.append(ifact.createPutStatic(cname, fname, IntegerType));
                cinitIlist.append(ifact.createPutStatic(cname, "temp_"+fname, IntegerType));
            }
            else
            {
                // Set up the references to point to the appropriate class's
                // (or interface's) state value
                if (StateHolder.isStateful(state.get(i).objName))
                {
                    fcname =
                        StateHolder.constructInterfaceName(state.get(i).objName);
                    invokeType = Constants.INVOKEINTERFACE;
                }
                else
                {
                    fcname = state.get(i).objName;
                    invokeType = Constants.INVOKEVIRTUAL;
                }
                ofname = state.get(i).name;
                fname = StateHolder.FIELD_NAME_PREFIX + state.get(i).name;
            }
            
            // Create a new method setState_X which will do as it says:
            // Update the temp state tempX with the given information.
            // If the new state value is "null", the policy has been violated.

            // Initialize a new ilist
            ilist = new InstructionList();

            // Change the state to the input value
            if (!isInstanceState)
            {
                // Place the first parameter (n) on the stack and update the
                // static field in this class
                ilist.append(ifact.ALOAD_0);
                ilist.append(ifact.createPutStatic(cname, "temp_" + fname, IntegerType));
            }
            else
            {
                // Place the object reference of the wrapper class onto the
                // stack, then the parameter n, then update the state contents
                ilist.append(ifact.ALOAD_1);
                ilist.append(ifact.ALOAD_0);
                ilist.append(ifact.createInvoke(
                        fcname, "set"+fname,
                        Type.VOID, new Type[] {IntegerType},
                        invokeType));
            }

            // Add a return (We're adding this now so we can jump
            // there when inserting an if command in a moment)
            ih = ilist.append(ifact.RETURN);

            // Add an if ("if (state == null) System.exit(0);")
            // We'll handle this by saying that if state != null, then
            // jump to our return point;
            // In between the branch and the instruction is the
            // System.exit(0) command
            instruction = ih.getInstruction();
            if (!isInstanceState)
            {
                ilist.insert(ih, ifact.createGetStatic(
                        cname, "temp_" + fname, IntegerType));
            }
            else
            {
                ilist.insert(ih, ifact.ALOAD_1);
//                ilist.insert(ih, ifact.createGetField(
//                        fcname, "temp_" + fname, IntegerType));
                ilist.insert(ih, ifact.createInvoke(
                        fcname, "getTemp"+fname,
                        IntegerType, new Type[] {},
                        invokeType));
            }
            //ilist.insert(instruction, ifact.createConstant(n));
            ilist.insert(ih, ifact.createBranchInstruction(
                    Constants.IFNONNULL, ih));
            // Integrate an error message
            ilist.insert(ih, createSystemOutput(
                    "\n" + "State " + ofname +
                    " violated the security policy." + "\n" +
                    "Aborting program.",
                    cpgen));
            // Create the exit call
            ilist.insert(ih, ifact.ICONST_M1);
            ilist.insert(ih, ifact.createInvoke(
                    "java.lang.System", "exit",
                    Type.VOID, new Type[] { Type.INT },
                    Constants.INVOKESTATIC));

            // Now create the "setter" method and add it
            mname = UniqueNameGen.findUniqueMethodName(cgen,
                    "setState_" + fname);
            if (!isInstanceState)
            {
                mg = new MethodGen(Constants.ACC_STATIC | Constants.ACC_PUBLIC,
                        Type.VOID, new Type[] {IntegerType}, new String[] {"iVal"},
                        mname, cname, ilist, cpgen);
            }
            else
            {
                mg = new MethodGen(Constants.ACC_STATIC | Constants.ACC_PUBLIC,
                        Type.VOID, new Type[] {IntegerType, Misc.constructType(fcname)},
                        new String[] {"iVal", "obj"}, mname, cname, ilist, cpgen);
            }
            mg.stripAttributes(true);
            mg.setMaxStack();
            mg.setMaxLocals();
            cgen.addMethod(mg.getMethod());
            ilist.dispose();
            
            setMethodName = mname;
            
            // We also need a "getter" method (Remove?)
            ilist = new InstructionList();
            
            // Grab the Integer value
            if (!isInstanceState)
            {
                ilist.append(ifact.createGetStatic(
                        cname, fname, IntegerType));
            }
            else
            {
                ilist.append(ifact.ALOAD_0);
                ilist.append(ifact.createInvoke(
                        fcname, "get"+fname,
                        IntegerType, new Type[] {},
                        invokeType));
            }
            // Cast as an int (No need to provide a full Integer wrapper class,
            // since we'll never be returning null as a value)
            ilist.append(ifact.createInvoke("java.lang.Integer", "intValue",
                    Type.INT, new Type[] {}, Constants.INVOKEVIRTUAL));
            ilist.append(ifact.IRETURN);
            
            // Now create the "getter" method and add it
            mname = UniqueNameGen.findUniqueMethodName(cgen,
                    "getState_" + fname);
            if (!isInstanceState)
            {
                mg = new MethodGen(Constants.ACC_STATIC | Constants.ACC_PUBLIC,
                        Type.INT, new Type[] {}, new String[] {},
                        mname, cname, ilist, cpgen);
            }
            else
            {
                mg = new MethodGen(Constants.ACC_STATIC | Constants.ACC_PUBLIC,
                        Type.INT, new Type[] {Misc.constructType(fcname)},
                        new String[] {"obj"}, mname, cname, ilist, cpgen);
            }
            mg.stripAttributes(true);
            mg.setMaxStack();
            mg.setMaxLocals();
            cgen.addMethod(mg.getMethod());
            ilist.dispose();
            
            getMethodName = mname;
            
            // Add the mapping
            stateMap.put(state.get(i).name,
                    new String[] {fname, setMethodName, getMethodName, fcname, state.get(i).objName});
            
        }
        
        // Now add the code for edges
        for (int i = 0; i < edge.size(); i++)
        {
            addPolicyEdge(edge.get(i), cname, cpgen, cgen, cinitIlist, stateMap,
                    new HashMap<String, Forall>(), IntegerType, invokeType);
            addPolicyEdgeCommit(edge.get(i), cname, cpgen, cgen,
                    stateMap, new HashMap<String, Forall>(), IntegerType, invokeType);
        }
        for (int i = 0; i < forall.size(); i++)
        {
            addPolicyForall(forall.get(i), cname, cpgen, cgen, cinitIlist, stateMap,
                    new HashMap<String, Forall>(), IntegerType, invokeType);
        }

        // Insert the new static initializer
        cinitIlist.append(ifact.RETURN);
        mg = new MethodGen(Constants.ACC_STATIC, Type.VOID, Type.NO_ARGS,
                new String[] {}, "<clinit>", cname,
                cinitIlist, cpgen);
        mg.stripAttributes(true);
        mg.setMaxStack();
        mg.setMaxLocals();
        cgen.addMethod(mg.getMethod());
        
        // Dispose of the instruction lists
        ilist.dispose();
        cinitIlist.dispose();
        
    }
    
    private void addPolicyForall(
            Forall forall,
            String cname,
            ConstantPoolGen cpgen,
            ClassGen cgen,
            InstructionList cinitIlist,
            HashMap<String, String[]> stateMap,
            HashMap<String, Forall> forallVar,
            Type IntegerType,
            short invokeType
            )
            throws Exception
    {
        // Adds edge methods based upon containing forall objects
        // Note that this method is recursive
        
        ArrayList<Forall> forallArrayList = forall.forall;
        ArrayList<Edge> edge = forall.edge;
        
        forallVar.put(forall.var, forall);

        for (int i = 0; i < forallArrayList.size(); i++)
        {
            addPolicyForall(forallArrayList.get(i), cname, cpgen, cgen, cinitIlist,
                    stateMap, forallVar, IntegerType, invokeType);
        }

        for (int i = 0; i < edge.size(); i++)
        {
            addPolicyEdge(edge.get(i), cname, cpgen, cgen, cinitIlist,
                    stateMap, forallVar, IntegerType, invokeType);
            addPolicyEdgeCommit(edge.get(i), cname, cpgen, cgen,
                    stateMap, forallVar, IntegerType, invokeType);
        }
        
        // Remove the forall variable from the hashmap
        forallVar.remove(forall.var);
        
    }
    
    private void addPolicyEdge(
            Edge edge,
            String cname,
            ConstantPoolGen cpgen,
            ClassGen cgen,
            InstructionList cinitIlist,
            HashMap<String, String[]> stateMap,
            HashMap<String, Forall> forallVar,
            Type IntegerType,
            short invokeType
            )
            throws Exception
    {
        // TODO: Make unique type-varname pairs (?)
        // Builds a edge method and inserts it into the class
        
        InstructionFactory ifact = new InstructionFactory(cpgen);
        
        // Build the instructions
        InstructionList ilist = new InstructionList();
        InstructionHandle ih = ilist.append(ifact.RETURN);
        // Generate code for each state for each of its participating edges
        // (i.e., node pairs)
        ArrayList<Nodes> nodes = edge.nodes;
        
        // Keep track of what forall variables have been computed and what their
        // corresponding local variables are
        HashMap<String, Integer> computedForall = new HashMap<String, Integer>();
        
        // Build a HashSet consisting of all the argument objects necessary
        // for us to do operations
        HashSet<String> argSet = new HashSet<String>();
        for (int i = 0; i < nodes.size(); i++)
        {
            if (!stateMap.get(nodes.get(i).state)[4].isEmpty())
            {
                argSet.add(stateMap.get(nodes.get(i).state)[3]);
            }
        }
        // Convert to an ArrayList to make our lives easier
        ArrayList<String> argArrayList = new ArrayList<String>();
        argArrayList.addAll(argSet);
        
        // Construct a list of types so that the method accepts objects
        // corresponding to each instance state variable
        Type[] typeArray = new Type[argArrayList.size()];
        for (int i = 0; i < argArrayList.size(); i++)
        {
            typeArray[i] = Misc.constructType(argArrayList.get(i));
        }

        // Construct the method gen
        MethodGen mgen = new MethodGen(
            Constants.ACC_STATIC | Constants.ACC_PUBLIC,
            Type.VOID, typeArray, null,
            "edge", cname, new InstructionList(), cpgen);
        ArrayList<LocalVariableGen> lvgen = new ArrayList<LocalVariableGen>();

        // Now build the following code (but only if there are edges
        // to merit it):
        // if (s==n1 && g==n2 && ..)
        //      ..
        // Insert before the last return so we
        // know where to direct our jumps
        // Build our giant if as follows:
        // Whenever ANY state value does not match the 'from'
        // attribute of a node pair, jump to the return
        // Otherwise keep falling through
        for (int i = 0; i < nodes.size(); i++)
        {
            // Is this an instance state?
            boolean isInstanceState =
                    !stateMap.get(nodes.get(i).state)[4].isEmpty();
            // Get the state value
            String fname = stateMap.get(nodes.get(i).state)[0];
            String fcname = stateMap.get(nodes.get(i).state)[3];
            String[] forallVarArray = new String[forallVar.size()];
            forallVar.keySet().toArray(forallVarArray);
            if (!isInstanceState)
            {
                ilist.insert(ih, ifact.createGetStatic(cname,
                        fname, IntegerType));
            }
            else
            {
                ilist.insert(ih, ifact.createLoad(
                        Misc.constructType(fcname), argArrayList.indexOf(fcname)));
                ilist.insert(ih, ifact.createInvoke(
                        fcname, "get"+fname,
                        IntegerType, new Type[] {},
                        invokeType));
            }
            ilist.insert(ih, ifact.createInvoke("java.lang.Integer",
                    "intValue", Type.INT, new Type[] {},
                    Constants.INVOKEVIRTUAL));
            // Check that field against the from value
            if (nodes.get(i).from.isInteger())
            {
                ilist.insert(ih, new PUSH(cpgen,
                        nodes.get(i).from.getInteger()));
                ilist.insert(ih,
                            ifact.createBranchInstruction(
                            Constants.IF_ICMPNE, ih));
            }
            else
            {
                // Must be a math expression
                LocalVariableGen newLvgen = mgen.addLocalVariable(
                        UniqueNameGen.findUniqueLocalVariableName(mgen, "lv"),
                        Type.INT, null, null);
                lvgen.add(newLvgen);
                ilist.insert(ih, ifact.createStore(Type.INT, newLvgen.getIndex()));
                ilist.insert(ih,
                        ExpressionCodeBuilder.buildFromCheck(
                        nodes.get(i), edge, forallVar,
                        computedForall, newLvgen.getIndex(), cgen, mgen));
                ilist.insert(ih, ifact.createBranchInstruction(Constants.GOTO, ih));
                InstructionHandle truePath = ilist.insert(ih, ifact.NOP);
                ilist.insert(truePath.getPrev().getPrev(),
                        ifact.createBranchInstruction(Constants.GOTO, truePath));
            }
//            else if (nodes.get(i).from.isVar(forallVarArray))
//            {
//                Forall compForall =
//                        forallVar.get(nodes.get(i).from.getString());
//                ilist.insert(ih, new PUSH(cpgen,
//                        compForall.from));
//                ilist.insert(ih,
//                        ifact.createBranchInstruction(
//                        Constants.IF_ICMPLT, ih));
//                if (!isInstanceState)
//                {
//                    ilist.insert(ih, ifact.createGetStatic(cname,
//                            fname, IntegerType));
//                }
//                else
//                {
//                    ilist.insert(ih, ifact.createLoad(
//                            Misc.constructType(fcname), argArrayList.indexOf(fcname)));
//                    ilist.insert(ih, ifact.createInvoke(
//                            fcname, "get"+fname,
//                            IntegerType, new Type[] {},
//                            invokeType));
//                }
//                ilist.insert(ih, ifact.createInvoke("java.lang.Integer",
//                        "intValue", Type.INT, new Type[] {},
//                        Constants.INVOKEVIRTUAL));
//                ilist.insert(ih, new PUSH(cpgen,
//                        compForall.to));
//                ilist.insert(ih,
//                        ifact.createBranchInstruction(
//                        Constants.IF_ICMPGT, ih));
//            }
//            else
//            {
//                ilist.insert(ih,
//                            ifact.createBranchInstruction(
//                            Constants.IFNE, ih));
//            }
        }

        // Now insert the code for updating the states
        for (int i = 0; i < nodes.size(); i++)
        {
            // Set the state accordingly
            // Is this an instance state?
            boolean isInstanceState =
                    !stateMap.get(nodes.get(i).state)[4].isEmpty();
            // Get the state value
            String fname = stateMap.get(nodes.get(i).state)[0];
            String fcname = stateMap.get(nodes.get(i).state)[3];
            // Push an Integer value or a null if we're
            // going to the "bad" state
            if (nodes.get(i).to.isInteger())
            {
                ilist.insert(ih, new PUSH(cpgen,
                        nodes.get(i).to.getInteger()));
                ilist.insert(ih, ifact.createInvoke("java.lang.Integer",
                        "valueOf", IntegerType, new Type[] {Type.INT},
                        Constants.INVOKESTATIC));
            }
            else if (nodes.get(i).to.isBadState())
            {
                ilist.insert(ih, ifact.ACONST_NULL);
            }
            else
            {
                // Math expression
                ilist.insert(ih, ExpressionCodeBuilder.buildTo(
                        nodes.get(i), edge, computedForall, cgen));
                ilist.insert(ih, ifact.createInvoke("java.lang.Integer",
                        "valueOf", IntegerType, new Type[] {Type.INT},
                        Constants.INVOKESTATIC));
            }
//            else if (nodes.get(i).to.getString().endsWith("+1"))
//            {
//                // For now, we only support increments
//                if (!isInstanceState)
//                {
//                    ilist.insert(ih, ifact.createGetStatic(cname,
//                            fname, IntegerType));
//                }
//                else
//                {
//                    ilist.insert(ih, ifact.createLoad(
//                            Misc.constructType(fcname), argArrayList.indexOf(fcname)));
//                    ilist.insert(ih, ifact.createInvoke(
//                            fcname, "get"+fname,
//                            IntegerType, new Type[] {},
//                            invokeType));
//                }
//                ilist.insert(ih, ifact.createInvoke("java.lang.Integer",
//                        "intValue", Type.INT, new Type[] {},
//                        Constants.INVOKEVIRTUAL));
//                ilist.insert(ih, new PUSH(cpgen, 1));
//                ilist.insert(ih, ifact.
//                        createBinaryOperation("+", Type.INT));
//                ilist.insert(ih, ifact.createInvoke("java.lang.Integer",
//                        "valueOf", IntegerType, new Type[] {Type.INT},
//                        Constants.INVOKESTATIC));
//            }

            if (!isInstanceState)
            {
                ilist.insert(ih, ifact.createInvoke(cname,
                        stateMap.get(nodes.get(i).state)[1], //setStateMethod
                        Type.VOID, new Type[] {IntegerType},
                        Constants.INVOKESTATIC));
            }
            else
            {
                ilist.insert(ih, ifact.createLoad(
                        Misc.constructType(fcname), argArrayList.indexOf(fcname)));
                ilist.insert(ih, ifact.createInvoke(cname,
                        stateMap.get(nodes.get(i).state)[1], //setStateMethod
                        Type.VOID, new Type[] {IntegerType, Misc.constructType(fcname)},
                        Constants.INVOKESTATIC));
            }

            // DEBUG: Uncomment the following to get state outputs at every
            // change
//            ilist.insert(ih, createSystemOutput(
//                    "\nState " + nodes.get(i).state + " = ", cpgen));
//            ilist.insert(ih, ifact.createFieldAccess(
//                "java.lang.System", "out",
//                new ObjectType("java.io.PrintStream"),
//                Constants.GETSTATIC));
//            ilist.insert(ih, ifact.createGetStatic(cname,
//                    stateMap.get(nodes.get(i).state)[0], IntegerType));
//            ilist.insert(ih, ifact.createInvoke("java.lang.Integer",
//                    "intValue", Type.INT, new Type[] {},
//                    Constants.INVOKEVIRTUAL));
//            ilist.insert(ih, ifact.createInvoke("java.io.PrintStream",
//                "print", Type.VOID, new Type[] { Type.INT },
//                Constants.INVOKEVIRTUAL));
        }

        // Now create the edge method and add it
        // NOTE: Currently assuming there is a valid distinct edge
        // name; might change and use a map
        String mname = UniqueNameGen.findUniqueMethodName(cgen,
                "edge_" + edge.name);
        mgen.setName(mname);
        mgen.setInstructionList(ilist);
        mgen.stripAttributes(true);
        mgen.setMaxStack();
        mgen.setMaxLocals();
        cgen.addMethod(mgen.getMethod());
        
        // Map the edge to its name
        edgeMap.put(edge, mgen.getMethod());
        
        ilist.dispose();
        
    }
    
    private void addPolicyEdgeCommit(
            Edge edge,
            String cname,
            ConstantPoolGen cpgen,
            ClassGen cgen,
            HashMap<String, String[]> stateMap,
            HashMap<String, Forall> forallVar,
            Type IntegerType,
            short invokeType
            )
            throws Exception
    {
        // Builds the corresponding commit for an edge method, which
        // sets all relevant state values to the values in their corresponding temps
        
        InstructionFactory ifact = new InstructionFactory(cpgen);
        
        // Build the instructions
        InstructionList ilist = new InstructionList();
        InstructionHandle ih = ilist.append(ifact.RETURN);
        // Get the nodes, so we know which states we're changing
        ArrayList<Nodes> nodes = edge.nodes;
        
        // Build a HashSet consisting of all the argument objects necessary
        // for us to do operations
        HashSet<String> argSet = new HashSet<String>();
        for (int i = 0; i < nodes.size(); i++)
        {
            if (!stateMap.get(nodes.get(i).state)[4].isEmpty())
            {
                argSet.add(stateMap.get(nodes.get(i).state)[3]);
            }
        }
        // Convert to an ArrayList to make our lives easier
        ArrayList<String> argArrayList = new ArrayList<String>();
        argArrayList.addAll(argSet);
        
        // Construct a list of types so that the method accepts objects
        // corresponding to each instance state variable
        Type[] typeArray = new Type[argArrayList.size()];
        for (int i = 0; i < argArrayList.size(); i++)
        {
            typeArray[i] = Misc.constructType(argArrayList.get(i));
        }

        // Construct the method gen
        MethodGen mgen = new MethodGen(
            Constants.ACC_STATIC | Constants.ACC_PUBLIC,
            Type.VOID, typeArray, null,
            "commit", cname, new InstructionList(), cpgen);

        // Now insert the code for updating the states
        for (int i = 0; i < nodes.size(); i++)
        {
            // Set the state accordingly
            // Is this an instance state?
            boolean isInstanceState =
                    !stateMap.get(nodes.get(i).state)[4].isEmpty();
            // Get the state value
            String fname = stateMap.get(nodes.get(i).state)[0];
            String fcname = stateMap.get(nodes.get(i).state)[3];

            // Copy over the values
            if (!isInstanceState)
            {
                ilist.insert(ih, ifact.createGetStatic(cname, "temp_"+fname, IntegerType));
                ilist.insert(ih, ifact.createPutStatic(cname, fname, IntegerType));
            }
            else
            {
                ilist.insert(ih, ifact.createLoad(
                        Misc.constructType(fcname), argArrayList.indexOf(fcname)));
                ilist.insert(ih, ifact.createInvoke(fcname, "commit"+fname,
                        Type.VOID, new Type[] {}, invokeType));
            }

        }

        // Now create the commit method and add it
        // NOTE: Currently assuming there is a valid distinct edge
        // name; might change and use a map
        String mname = UniqueNameGen.findUniqueMethodName(cgen,
                "commit_" + edge.name);
        mgen.setName(mname);
        mgen.setInstructionList(ilist);
        mgen.stripAttributes(true);
        mgen.setMaxStack();
        mgen.setMaxLocals();
        cgen.addMethod(mgen.getMethod());
        
        ilist.dispose();
        
    }
    
    public InstructionList createSystemOutput(
            String text, ConstantPoolGen cpgen)
    {
        // Convenience method for generating output statements for insertion
        InstructionList ilist = new InstructionList();
        InstructionFactory ifact = new InstructionFactory(cpgen);
        ilist.append(ifact.createFieldAccess(
            "java.lang.System", "out",
            new ObjectType("java.io.PrintStream"),
            Constants.GETSTATIC));
        ilist.append(new PUSH(cpgen, text));
        ilist.append(ifact.createInvoke("java.io.PrintStream",
            "print", Type.VOID, new Type[] { Type.STRING },
            Constants.INVOKEVIRTUAL));
        return ilist;
    }
    
}
